
package geniusweb.javavenv;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.logging.Level;

import org.apache.commons.io.FilenameUtils;

import com.fasterxml.jackson.databind.ObjectMapper;

import geniusweb.party.Capabilities;
import tudelft.utilities.logging.ReportToLogger;

/**
 * Factory that can create instances of a specific party with connector running
 * on a {@link PythonVenv}. The idea is that you create a factory for each targz
 * file, so that you can create many party instances quickly without need to
 * re-create a new venv
 */
public class PythonPartyFactory {

	private static final int PARTY_MAX_INSTALL_TIME = 60000; // ms
	private static final int PIP_MAX_UPGRADE_TIME = 30000; // ms
	private final ObjectMapper jackson = new ObjectMapper();
	private final File targz;
	private final transient PythonVenv venv;
	private final Capabilities capabilities;
	private final String descriptino;
	private ReportToLogger logger;

	/**
	 * 
	 * @param targz the tar.gz fiel containing the party to be loaded.
	 * @throws IOException          if party can not be loaded from file
	 * @throws PythonError          if party has a error in the python code
	 * @throws InterruptedException if connection with party is interrupted
	 */
	public PythonPartyFactory(File targz)
			throws IOException, InterruptedException, PythonError {
		if (!targz.exists())
			throw new IllegalArgumentException("File not found: " + targz);
		if (!targz.canRead())
			throw new IllegalArgumentException(
					"File can not be read: " + targz);
		this.logger = new ReportToLogger(
				"py_" + FilenameUtils.getBaseName(targz.getName()));
		this.targz = targz;
		this.venv = getVenv();
		this.capabilities = fetchCapabilities();
		this.descriptino = fetchDescription();
	}

	/**
	 * 
	 * @return the {@link Capabilities} of this party
	 */
	public Capabilities getCapabilites() {
		return capabilities;
	}

	/**
	 * 
	 * @return the Description of this party
	 */
	public String getDescription() {
		return descriptino;
	}

	/**
	 * @return a Pipe to an instance of this party.
	 * @throws IOException if
	 */
	public PartyProxy getInstance() throws IOException {
		Pipe pipe = venv.pipe(Arrays.asList("-c",
				"from geniusweb.partystdio.PartyStdIo import PartyStdIo; import party; PartyStdIo(party.party()).run()"),
				500);
		return new PartyProxy(pipe, getCapabilites(), getDescription(), logger);
	}

	public void close() {
		venv.remove();
	}

	protected PythonVenv getVenv()
			throws IOException, InterruptedException, PythonError {
		String name = targz.getName().replaceAll("\\W", "");
		PythonVenv env = new PythonVenv(name, logger);
		env.call("-m pip install --upgrade pip", PIP_MAX_UPGRADE_TIME);
		env.call(Arrays.asList("-m", "pip", "install", targz.getAbsolutePath()),
				PARTY_MAX_INSTALL_TIME);

		return env;
	}

	private Capabilities fetchCapabilities()
			throws InterruptedException, IOException, PythonError {
		String capstr = venv.call(Arrays.asList("-c",
				"from geniusweb.partystdio.PartyStdIo import PartyStdIo; import party; PartyStdIo(party.party()).capabilities()"),
				500);
		logger.log(Level.INFO, "Capabilities=" + capstr);
		return jackson.readValue(capstr, Capabilities.class);
	}

	private String fetchDescription()
			throws InterruptedException, IOException, PythonError {
		String capstr = venv.call(Arrays.asList("-c",
				"from geniusweb.partystdio.PartyStdIo import PartyStdIo; import party; PartyStdIo(party.party()).description()"),
				500);
		logger.log(Level.INFO, "Description=" + capstr);
		return capstr;
	}

}